from __future__ import annotations

import warnings
from logging import getLogger
from typing import Any, Literal

from lightgbm import LGBMModel
from typing_extensions import Self

from .._early_stopping_callback import DartEarlyStoppingCallback, EarlyStoppingCallback
from ._base import EstimatorWrapperBase

LOG = getLogger(__name__)


class LGBMDartEarlyStoppingSimpleWrapper(EstimatorWrapperBase[LGBMModel]):
    """A simple wrapper for dart LGBMModel that returns the best model after early stopping."""

    def __init__(
        self,
        estimator: LGBMModel,
        *,
        method: Literal["save", "refit", "none"] = "save",
        metric_idx: int = -1,
    ) -> None:
        """A simple wrapper for dart LGBMModel that returns the best model after early stopping.

        Parameters
        ----------
        estimator : LGBMModel
            The estimator to be wrapped.
        method : Literal[&quot;save&quot;, &quot;refit&quot;, &quot;none&quot;], optional
            The method to return the best model, by default &quot;save&quot;
            "save": Save the best model by deepcopying the estimator and return the best model.
            "refit": Refit the estimator with the best iteration and return the refitted estimator.
            "none": Do nothing and return the original estimator.
        metric_idx : int, optional
            The index of the metric to be used for early stopping, by default 0
        """
        self.method = method
        self.metric_idx = metric_idx
        super().__init__(estimator)

    def fit(self, X, y, **fit_params: Any) -> Self:  # type: ignore
        early_stopping_callback_candidates = [
            callback
            for callback in fit_params.get("callbacks", [])
            if isinstance(callback, EarlyStoppingCallback)
        ]
        if (
            self.estimator.get_params()["boosting_type"] != "dart"
            or (len(early_stopping_callback_candidates) == 0)
            or (self.method == "none")
        ):
            self.estimator.fit(X, y, **fit_params)
            return self

        if len(early_stopping_callback_candidates) > 1:
            warnings.warn(
                "Multiple EarlyStoppingCallback objects are found. "
                "Only the first one is used.",
                UserWarning,
            )
        early_stopping_callback = early_stopping_callback_candidates[0]

        if self.method in ["refit", "refit_like_save"]:
            self.estimator.fit(X, y, **fit_params)
            LOG.debug(f"best_iter: {early_stopping_callback.best_iter}")
            self.estimator.set_params(
                n_estimators=early_stopping_callback.best_iter[self.metric_idx]
                + 1
                + (1 if self.method == "refit_like_save" else 0)
            )
            self.estimator.fit(X, y, **fit_params)

        elif self.method == "save":
            if not isinstance(early_stopping_callback, DartEarlyStoppingCallback):
                raise ValueError(
                    "EarlyStoppingCallback is not DartEarlyStoppingCallback. "
                    f"Got {type(early_stopping_callback)}"
                )
            self.estimator.fit(X, y, **fit_params)
            self.estimator._Booster = early_stopping_callback.best_model[
                self.metric_idx
            ]

        else:
            raise ValueError(f"Unknown method: {self.method}")
        return self



__version__ = "0.1.9"
from ._base import CallbackBase
from ._early_stopping_callback import DartEarlyStoppingCallback, EarlyStoppingCallback
from ._tqdm_callback import ProgressBarCallback
from .sklearn import (
    EstimatorWrapperBase,
    LGBMDartEarlyStoppingEstimator,
    LGBMDartEarlyStoppingSimpleWrapper,
    LGBMEarlyStoppingEstimator,
)

__all__ = [
    "CallbackBase",
    "ProgressBarCallback",
    "EarlyStoppingCallback",
    "DartEarlyStoppingCallback",
    "LGBMEarlyStoppingEstimator",
    "LGBMDartEarlyStoppingEstimator",
    "EstimatorWrapperBase",
    "LGBMDartEarlyStoppingSimpleWrapper",
]



from __future__ import annotations

import pickle  # nosec
from functools import partial
from typing import Any, Callable, List, Tuple, TypeVar, Union

from lightgbm.basic import _ConfigAliases, _log_info, _log_warning
from lightgbm.callback import CallbackEnv, EarlyStopException, _format_eval_result

_LGBM_BoosterEvalMethodResultType = Tuple[str, str, float, bool]
_ListOfEvalResultTuples = Union[
    List[_LGBM_BoosterEvalMethodResultType], List[Tuple[str, str, float, bool, float]]
]

TObj = TypeVar("TObj", bound=Any)


def _complete_deepcopy(obj: TObj) -> TObj:
    return pickle.loads(pickle.dumps(obj))  # nosec


def _gt_delta(curr_score: float, best_score: float, delta: float) -> bool:
    return curr_score > best_score + delta


def _lt_delta(curr_score: float, best_score: float, delta: float) -> bool:
    return curr_score < best_score - delta


def _is_train_set(ds_name: str, eval_name: str, train_name: str) -> bool:
    return (ds_name == "cv_agg" and eval_name == "train") or ds_name == train_name


class EarlyStoppingCallback:
    """A callback that activates early stopping.

    Activates early stopping.
    The model will train until the validation score doesn't improve by at least ``min_delta``.
    Validation score needs to improve at least every ``stopping_rounds`` round(s)
    to continue training.
    Requires at least one validation data and one metric.
    If there's more than one, will check all of them. But the training data is ignored anyway.
    To check only the first metric set ``first_metric_only`` to True.
    The index of iteration that has the best performance will be saved in the ``best_iteration``
    attribute of a model.

    Compared to the official implementation, the best_iteration information
    is retained even when using dart.

    Parameters
    ----------
    stopping_rounds : int
        The possible number of rounds without the trend occurrence.
    first_metric_only : bool, optional (default=False)
        Whether to use only the first metric for early stopping.
    verbose : bool, optional (default=True)
        Whether to log message with early stopping information.
        By default, standard output resource is used.
        Use ``register_logger()`` function to register a custom logger.
    min_delta : float or list of float, optional (default=0.0)
        Minimum improvement in score to keep training.
        If float, this single value is used for all metrics.
        If list, its length should match the total number of metrics."""

    def __init__(
        self,
        stopping_rounds: int,
        first_metric_only: bool = False,
        verbose: bool = True,
        min_delta: float | list[float] = 0.0,
    ) -> None:
        self.order = 30
        self.before_iteration = False

        self.stopping_rounds = stopping_rounds
        self.first_metric_only = first_metric_only
        self.verbose = verbose
        self.min_delta = min_delta

        self.enabled = True
        self._reset_storages()

    def _reset_storages(self) -> None:
        self.best_score: list[float] = []
        self.best_iter: list[int] = []
        self.best_score_list: list[_ListOfEvalResultTuples] = []
        self.cmp_op: list[Callable[[float, float], bool]] = []
        self.first_metric = ""

    def _init(self, env: CallbackEnv) -> None:
        is_dart = any(
            env.params.get(alias, "") == "dart"
            for alias in _ConfigAliases.get("boosting")
        )
        self.is_dart = is_dart
        only_train_set = len(env.evaluation_result_list) == 1 and _is_train_set(
            ds_name=env.evaluation_result_list[0][0],
            eval_name=env.evaluation_result_list[0][1].split(" ")[0],
            train_name=env.model._train_data_name,
        )
        self.enabled = (
            not only_train_set
        )  # not is_dart and not only_train_set # NOTE: added by 34j
        if not self.enabled:
            if is_dart:
                _log_warning("Early stopping is not available in dart mode")
            elif only_train_set:
                _log_warning("Only training set found, disabling early stopping.")
            return
        if not env.evaluation_result_list:
            raise ValueError(
                "For early stopping, "
                "at least one dataset and eval metric is required for evaluation"
            )

        if self.stopping_rounds <= 0:
            raise ValueError("stopping_rounds should be greater than zero.")

        if self.verbose:
            _log_info(
                f"Training until validation scores don't improve for {self.stopping_rounds} rounds"
            )

        self._reset_storages()

        n_metrics = len({m[1] for m in env.evaluation_result_list})
        n_datasets = len(env.evaluation_result_list) // n_metrics
        if isinstance(self.min_delta, list):
            if not all(t >= 0 for t in self.min_delta):
                raise ValueError(
                    "Values for early stopping min_delta must be non-negative."
                )
            if len(self.min_delta) == 0:
                if self.verbose:
                    _log_info("Disabling min_delta for early stopping.")
                deltas = [0.0] * n_datasets * n_metrics
            elif len(self.min_delta) == 1:
                if self.verbose:
                    _log_info(
                        f"Using {self.min_delta[0]} as min_delta for all metrics."
                    )
                deltas = self.min_delta * n_datasets * n_metrics
            else:
                if len(self.min_delta) != n_metrics:
                    raise ValueError(
                        "Must provide a single value for min_delta or as many as metrics."
                    )
                if self.first_metric_only and self.verbose:
                    _log_info(
                        f"Using only {self.min_delta[0]} as early stopping min_delta."
                    )
                deltas = self.min_delta * n_datasets
        else:
            if self.min_delta < 0:
                raise ValueError("Early stopping min_delta must be non-negative.")
            if (
                self.min_delta > 0
                and n_metrics > 1
                and not self.first_metric_only
                and self.verbose
            ):
                _log_info(f"Using {self.min_delta} as min_delta for all metrics.")
            deltas = [self.min_delta] * n_datasets * n_metrics

        # split is needed for "<dataset type> <metric>" case (e.g. "train l1")
        self.first_metric = env.evaluation_result_list[0][1].split(" ")[-1]
        for eval_ret, delta in zip(env.evaluation_result_list, deltas):
            self.best_iter.append(0)
            if eval_ret[3]:  # greater is better
                self.best_score.append(float("-inf"))
                self.cmp_op.append(partial(_gt_delta, delta=delta))
            else:
                self.best_score.append(float("inf"))
                self.cmp_op.append(partial(_lt_delta, delta=delta))

    def _final_iteration_check(
        self, env: CallbackEnv, eval_name_splitted: list[str], i: int
    ) -> None:
        if env.iteration == env.end_iteration - 1:
            if self.verbose:
                best_score_str = "\t".join(
                    [
                        _format_eval_result(x, show_stdv=True)
                        for x in self.best_score_list[i]
                    ]
                )
                _log_info(
                    "Did not meet early stopping. "
                    f"Best iteration is:\n[{self.best_iter[i] + 1}]\t{best_score_str}"
                )
                if self.first_metric_only:
                    _log_info(f"Evaluated only: {eval_name_splitted[-1]}")
            if self.is_dart:  # NOTE: added by 34j
                # return current iter
                raise EarlyStopException(env.iteration, self.best_score_list[i])
            raise EarlyStopException(self.best_iter[i], self.best_score_list[i])

    # NOTE: added by 34j
    def on_best_score_updated(self, env: CallbackEnv, i: int) -> None:
        pass

    def on_best_score_not_updated(self, env: CallbackEnv, i: int) -> None:
        pass

    def on_best_score_update_started(self, env: CallbackEnv, i: int) -> None:
        pass

    def on_best_score_update_stopped(self, env: CallbackEnv, i: int) -> None:
        pass

    def __call__(self, env: CallbackEnv) -> None:
        if env.iteration == env.begin_iteration:
            self._init(env)
        if not self.enabled:
            return

        # self.best_score_list is initialized to an empty list
        first_time_updating_best_score_list = self.best_score_list == []
        for i in range(len(env.evaluation_result_list)):
            score = env.evaluation_result_list[i][2]
            if first_time_updating_best_score_list or self.cmp_op[i](
                score, self.best_score[i]
            ):
                # NOTE: added by 34j
                if self.best_iter[i] != env.iteration - 1:
                    self.on_best_score_update_started(env, i)
                self.on_best_score_updated(env, i)

                self.best_score[i] = score
                self.best_iter[i] = env.iteration
                if first_time_updating_best_score_list:
                    self.best_score_list.append(env.evaluation_result_list)
                else:
                    self.best_score_list[i] = env.evaluation_result_list
            else:
                # NOTE: added by 34j
                if self.best_iter[i] == env.iteration - 1:
                    self.on_best_score_update_stopped(env, i)
                self.on_best_score_not_updated(env, i)
            # split is needed for "<dataset type> <metric>" case (e.g. "train l1")
            eval_name_splitted = env.evaluation_result_list[i][1].split(" ")
            if self.first_metric_only and self.first_metric != eval_name_splitted[-1]:
                continue  # use only the first metric for early stopping
            if _is_train_set(
                env.evaluation_result_list[i][0],
                eval_name_splitted[0],
                env.model._train_data_name,
            ):
                continue  # train data for lgb.cv or sklearn wrapper (underlying lgb.train)
            elif env.iteration - self.best_iter[i] >= self.stopping_rounds:
                if self.verbose:
                    eval_result_str = "\t".join(
                        [
                            _format_eval_result(x, show_stdv=True)
                            for x in self.best_score_list[i]
                        ]
                    )
                    _log_info(
                        f"Early stopping, best iteration is:\n[{self.best_iter[i] + 1}]\t"
                        f"{eval_result_str}"
                    )
                    if self.first_metric_only:
                        _log_info(f"Evaluated only: {eval_name_splitted[-1]}")
                if self.is_dart:  # NOTE: added by 34j
                    # return current iter
                    raise EarlyStopException(env.iteration, self.best_score_list[i])
                raise EarlyStopException(self.best_iter[i], self.best_score_list[i])
            self._final_iteration_check(env, eval_name_splitted, i)


class DartEarlyStoppingCallback(EarlyStoppingCallback):
    """A callback that activates early stopping.

    Activates early stopping.
    The model will train until the validation score doesn't improve by at least ``min_delta``.
    Validation score needs to improve at least every ``stopping_rounds`` round(s)
    to continue training.
    Requires at least one validation data and one metric.
    If there's more than one, will check all of them. But the training data is ignored anyway.
    To check only the first metric set ``first_metric_only`` to True.
    The index of iteration that has the best performance will be saved in the ``best_iteration``
    attribute of a model.

    When using dart, the model is copied and retained using pickle at iterations
    where scores no longer improve. Eventually, the number of iterations of the ``best_model``
    should be ``max(best_iteration[i] + 2, num_boost_round/n_iter/n_estimators)``.
    (``max(best_iteration[i] + 1, num_boost_round/n_iter/n_estimators)`` in case of normal early stopping.)

    Parameters
    ----------
    stopping_rounds : int
        The possible number of rounds without the trend occurrence.
    first_metric_only : bool, optional (default=False)
        Whether to use only the first metric for early stopping.
    verbose : bool, optional (default=True)
        Whether to log message with early stopping information.
        By default, standard output resource is used.
        Use ``register_logger()`` function to register a custom logger.
    min_delta : float or list of float, optional (default=0.0)
        Minimum improvement in score to keep training.
        If float, this single value is used for all metrics.
        If list, its length should match the total number of metrics."""

    def _reset_storages(self) -> None:
        self.best_model: list[Any] = []
        return super()._reset_storages()

    def _save_model(self, env: CallbackEnv, i: int) -> None:
        # _log_info(f"Saving best model for metric {i}.")
        self.best_model.extend([None] * (i + 1 - len(self.best_model)))
        self.best_model[i] = _complete_deepcopy(env.model)

    def on_best_score_update_stopped(self, env: CallbackEnv, i: int) -> None:
        if not self.is_dart:
            return

        self._save_model(env, i)

    def _final_iteration_check(
        self, env: CallbackEnv, eval_name_splitted: list[str], i: int
    ) -> None:
        if not self.is_dart:
            return super()._final_iteration_check(env, eval_name_splitted, i)

        if env.iteration == env.end_iteration - 1:
            self._save_model(env, i)
        return super()._final_iteration_check(env, eval_name_splitted, i)



from __future__ import annotations

import warnings
from typing import Any, Callable, Literal

import tqdm
from lightgbm import LGBMModel, log_evaluation
from lightgbm.callback import CallbackEnv
from numpy.typing import NDArray
from sklearn.model_selection import train_test_split
from typing_extensions import Self

from lightgbm_callbacks import DartEarlyStoppingCallback

from .._tqdm_callback import ProgressBarCallback
from ._base import EstimatorWrapperBase
from ._dart_early_stopping_wrapper import LGBMDartEarlyStoppingSimpleWrapper


class LGBMEarlyStoppingEstimator(EstimatorWrapperBase[LGBMModel]):
    """LightGBM wrapper that does early stopping with sklearn.train_test_split."""

    def __init__(
        self,
        estimator: LGBMModel,
        *,
        early_stopping_factory: Callable[
            [int, bool, bool, float | list[float]], Callable[[CallbackEnv], None]
        ]
        | Callable[
            [int, bool, bool], Callable[[CallbackEnv], None]
        ] = DartEarlyStoppingCallback,
        stopping_rounds: int | None = None,
        first_metric_only: bool = False,
        verbose: bool = False,
        min_delta: float | None = None,
        eval_metric: str
        | Callable[[NDArray, NDArray], tuple[str, float, bool]]
        | None = None,
        test_size: float | int | None = None,
        train_size: float | int | None = None,
        random_state: int = 0,
        shuffle: bool = True,
        stratify: bool = False,
        split_enabled: bool = True,
        tqdm_cls: Literal[
            "auto",
            "autonotebook",
            "std",
            "notebook",
            "asyncio",
            "keras",
            "dask",
            "tk",
            "gui",
            "rich",
            "contrib.slack",
            "contrib.discord",
            "contrib.telegram",
            "contrib.bells",
        ]
        | type[tqdm.std.tqdm]
        | None = None,
        tqdm_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """LightGBM wrapper that does early stopping with sklearn.train_test_split.

        Parameters
        ----------
        estimator : LGBMModel
            Scikit-learn API LightGBM estimator.
        stopping_rounds : int
            The possible number of rounds without the trend occurrence.
            Alias: ``n_iter_no_change``
        early_stopping_factory : Callable[[int, bool, bool, float | list[float]],
        Callable[CallbackEnv, None]] | Callable[[int, bool, bool], Callable[CallbackEnv, None]],
        optional
            Factory function that returns a callback function, by default DartEarlyStoppingCallback
        first_metric_only : bool, optional
            Whether to use only the first metric for early stopping or use all of them, by default False
        min_delta : float, optional
            Minimum delta to be considered as an actual change, by default 0.0
            Alias: ``tol``
        verbose : bool, optional
            Whether to print message about early stopping, by default False
        eval_metric : str | Callable[[NDArray, NDArray], tuple[str, float, bool]], optional
            Evaluation metric, by default "rmse"
        test_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. If ``train_size`` is also None, it will
            be set to 0.25.
            Alias: ``validation_fraction``
        train_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the train split. If
            int, represents the absolute number of train samples. If None,
            the value is automatically set to the complement of the test size.
            Alias: ``train_fraction``
        random_state : int, RandomState instance or None, default=None
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.
        shuffle : bool, optional
            Whether or not to shuffle the data before splitting. If shuffle=False
            then stratify must be None. by default True
        stratify : bool, optional
            Whether or not stratify the data before splitting. If stratify=True,
            y must be categorical. by default False
        split_enabled : bool, optional
            Whether to use train_test_split or not, by default True
        tqdm_cls : Literal['auto', 'autonotebook', 'std', 'notebook', 'asyncio',
        'keras', 'dask', 'tk', 'gui', 'rich', 'contrib.slack', 'contrib.discord',
        'contrib.telegram', 'contrib.bells'] or type[tqdm.std.tqdm] or None, optional
            The tqdm class or module name, by default 'auto'
        tqdm_kwargs : dict[str, Any] or None, optional
            The keyword arguments passed to the tqdm class initializer, by default None
        **kwargs : Any
            Other parameters passed to the estimator.
        """
        self.estimator = estimator
        self.early_stopping_factory = early_stopping_factory
        self.stopping_rounds = stopping_rounds
        self.first_metric_only = first_metric_only
        self.verbose = verbose
        self.min_delta = min_delta
        self.eval_metric = eval_metric
        self.test_size = test_size
        self.train_size = train_size
        self.random_state = random_state
        self.shuffle = shuffle
        self.stratify = stratify
        self.split_enabled = split_enabled
        self.tqdm_cls = tqdm_cls
        self.tqdm_kwargs = tqdm_kwargs
        self.kwargs = kwargs
        for key, value in kwargs.items():
            if key not in [
                "validation_fraction",
                "train_fraction",
                "tol",
                "n_iter_no_change",
            ]:
                warnings.warn(f"Unknown parameter: {key}: {value}")

    def fit(self, X, y=None, **fit_params) -> Self:  # type: ignore
        """Fit the model according to the given training data."""
        if not self.split_enabled:
            self.estimator.fit(X, y, **fit_params)
            return self
        X_train, X_test, y_train, y_test = train_test_split(
            X,
            y,
            test_size=self.test_size or self.kwargs.get("validation_fraction", None),
            train_size=self.train_size or self.kwargs.get("train_fraction", None),
            random_state=self.random_state,
            shuffle=self.shuffle,
            stratify=y if self.stratify else None,
        )
        fit_params["eval_set"] = [(X_train, y_train), (X_test, y_test)]
        if self.eval_metric is not None:
            fit_params["eval_metric"] = self.eval_metric
        stopping_rounds = self.stopping_rounds or self.kwargs.get(
            "n_iter_no_change", None
        )
        if stopping_rounds is not None:
            try:
                early_stopping = self.early_stopping_factory(
                    stopping_rounds,
                    self.first_metric_only,
                    self.verbose,
                    self.min_delta or self.kwargs.get("tol", 0.0),  # type: ignore
                )
            except TypeError:
                early_stopping = self.early_stopping_factory(
                    stopping_rounds, self.first_metric_only, self.verbose  # type: ignore
                )
            fit_params["callbacks"] = [
                early_stopping,
                log_evaluation(self.verbose),
            ] + fit_params.get("callbacks", [])
            for callback in fit_params["callbacks"]:
                if isinstance(callback, ProgressBarCallback):
                    callback.early_stopping_callback = early_stopping
            if self.tqdm_cls is not None:
                fit_params["callbacks"].append(
                    ProgressBarCallback(
                        tqdm_cls=self.tqdm_cls,
                        early_stopping_callback=early_stopping,
                        **(self.tqdm_kwargs or {}),
                    )
                )
        self.estimator.fit(
            X_train,
            y_train,
            **fit_params,
        )
        return self


class LGBMDartEarlyStoppingEstimator(LGBMEarlyStoppingEstimator):
    """LightGBM wrapper that does early stopping with sklearn.train_test_split
    and uses ``LGBMDartEarlyStoppingSimpleWrapper`` to support early stopping in dart mode.
    """

    def __init__(
        self,
        estimator: LGBMModel,
        *,
        early_stopping_factory: Callable[
            [int, bool, bool, float | list[float]], Callable[[CallbackEnv], None]
        ]
        | Callable[
            [int, bool, bool], Callable[[CallbackEnv], None]
        ] = DartEarlyStoppingCallback,
        stopping_rounds: int | None = None,
        first_metric_only: bool = False,
        verbose: bool = False,
        min_delta: float | None = None,
        eval_metric: str
        | Callable[[NDArray, NDArray], tuple[str, float, bool]]
        | None = None,
        test_size: float | int | None = None,
        train_size: float | int | None = None,
        random_state: int = 0,
        shuffle: bool = True,
        stratify: bool = False,
        split_enabled: bool = True,
        dart_early_stopping_method: Literal["save", "refit", "none"] = "save",
        metric_idx: int = -1,
        tqdm_cls: Literal[
            "auto",
            "autonotebook",
            "std",
            "notebook",
            "asyncio",
            "keras",
            "dask",
            "tk",
            "gui",
            "rich",
            "contrib.slack",
            "contrib.discord",
            "contrib.telegram",
            "contrib.bells",
        ]
        | type[tqdm.std.tqdm]
        | None = None,
        tqdm_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """LightGBM wrapper that does early stopping with sklearn.train_test_split
        and uses ``LGBMDartEarlyStoppingSimpleWrapper`` to support early stopping in dart mode.

        Parameters
        ----------
        estimator : LGBMModel
            Scikit-learn API LightGBM estimator.
        stopping_rounds : int
            The possible number of rounds without the trend occurrence.
            Alias: ``n_iter_no_change``
        early_stopping_factory : Callable[[int, bool, bool, float | list[float]],
        Callable[CallbackEnv, None]] | Callable[[int, bool, bool], Callable[CallbackEnv, None]],
        optional
            Factory function that returns a callback function, by default DartEarlyStoppingCallback
        first_metric_only : bool, optional
            Whether to use only the first metric for early stopping or use all of them, by default False
        min_delta : float, optional
            Minimum delta to be considered as an actual change, by default 0.0
            Alias: ``tol``
        verbose : bool, optional
            Whether to print message about early stopping, by default False
        eval_metric : str | Callable[[NDArray, NDArray], tuple[str, float, bool]], optional
            Evaluation metric, by default "rmse"
        test_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. If ``train_size`` is also None, it will
            be set to 0.25.
            Alias: ``validation_fraction``
        train_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the train split. If
            int, represents the absolute number of train samples. If None,
            the value is automatically set to the complement of the test size.
            Alias: ``train_fraction``
        random_state : int, RandomState instance or None, default=None
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.
        shuffle : bool, optional
            Whether or not to shuffle the data before splitting. If shuffle=False
            then stratify must be None. by default True
        stratify : bool, optional
            Whether or not stratify the data before splitting. If stratify=True,
            y must be categorical. by default False
        split_enabled : bool, optional
            Whether to use train_test_split or not, by default True
        dart_early_stopping_method : Literal["save", "refit", "none"], optional
            Method to use for early stopping, by default "save"
        metric_idx : int, optional
            Index of the metric to use for early stopping, by default 0
        tqdm_cls : Literal['auto', 'autonotebook', 'std', 'notebook', 'asyncio',
        'keras', 'dask', 'tk', 'gui', 'rich', 'contrib.slack', 'contrib.discord',
        'contrib.telegram', 'contrib.bells'] or type[tqdm.std.tqdm] or None, optional
            The tqdm class or module name, by default 'auto'
        tqdm_kwargs : dict[str, Any] or None, optional
            The keyword arguments passed to the tqdm class initializer, by default None
        **kwargs : Any
            Other parameters passed to the estimator.
        """
        super().__init__(
            estimator,
            early_stopping_factory=early_stopping_factory,
            stopping_rounds=stopping_rounds,
            first_metric_only=first_metric_only,
            verbose=verbose,
            min_delta=min_delta,
            eval_metric=eval_metric,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify,
            split_enabled=split_enabled,
            tqdm_cls=tqdm_cls,
            tqdm_kwargs=tqdm_kwargs,
            **kwargs,
        )
        self.dart_early_stopping_method = dart_early_stopping_method
        self.metric_idx = metric_idx

    def fit(self, X, y=None, **fit_params) -> Self:  # type: ignore
        self.estimator = LGBMDartEarlyStoppingSimpleWrapper(
            self.estimator,
            method=self.dart_early_stopping_method,
            metric_idx=self.metric_idx,
        )
        super().fit(X, y, **fit_params)
        self.estimator = self.estimator.estimator
        return self



from __future__ import annotations

from lightgbm.callback import CallbackEnv


class CallbackBase:
    def _init(self, env: CallbackEnv) -> None:
        pass  # pragma: no cover

    def __call__(self, env: CallbackEnv) -> None:
        if env.iteration == env.begin_iteration:
            self._init(env)



from __future__ import annotations

import importlib
import warnings
from collections import OrderedDict
from typing import Any, Literal

import tqdm
from lightgbm.callback import CallbackEnv

from ._base import CallbackBase


class ProgressBarCallback(CallbackBase):
    tqdm_cls: type[tqdm.std.tqdm] | None
    pbar: tqdm.std.tqdm | None

    def __init__(
        self,
        tqdm_cls: Literal[
            "auto",
            "autonotebook",
            "std",
            "notebook",
            "asyncio",
            "keras",
            "dask",
            "tk",
            "gui",
            "rich",
            "contrib.slack",
            "contrib.discord",
            "contrib.telegram",
            "contrib.bells",
        ]
        | type[tqdm.std.tqdm]
        | None = "auto",
        early_stopping_callback: Any | None = None,
        **tqdm_kwargs: Any,
    ) -> None:
        """Progress bar callback for LightGBM training.

        Parameters
        ----------
        tqdm_cls : Literal['auto', 'autonotebook', 'std', 'notebook', 'asyncio',
        'keras', 'dask', 'tk', 'gui', 'rich', 'contrib.slack', 'contrib.discord',
        'contrib.telegram', 'contrib.bells'] or type[tqdm.std.tqdm] or None, optional
            The tqdm class or module name, by default 'auto'
        early_stopping_callback : Any | None, optional
            The early stopping callback, by default None
        **tqdm_kwargs : Any
            The keyword arguments passed to the tqdm class initializer

            .. rubric:: Example

            .. code-block:: python
                early_stopping_callback = early_stopping(stopping_rounds=50)
                callbacks = [
                    early_stopping_callback,
                    ProgressBarCallback(early_stopping_callback=early_stopping_callback)
                ]
                estimator.fit(X_train, y_train, eval_set=[(X_test, y_test)], callbacks=callbacks)
        """
        self.order = 40
        self.before_iteration = False
        if isinstance(tqdm_cls, str):
            tqdm_module = importlib.import_module(f"tqdm.{tqdm_cls}")
            self.tqdm_cls = getattr(tqdm_module, "tqdm")
        else:
            self.tqdm_cls = tqdm_cls
        self.early_stopping_callback = early_stopping_callback
        self.tqdm_kwargs = tqdm_kwargs
        if "total" in tqdm_kwargs:
            warnings.warn("'total' in tqdm_kwargs is ignored.", UserWarning)
        self.pbar = None

    def _init(self, env: CallbackEnv) -> None:
        # create pbar on first call
        if self.tqdm_cls is None:
            return
        tqdm_kwargs = self.tqdm_kwargs.copy()
        tqdm_kwargs["total"] = env.end_iteration - env.begin_iteration
        self.pbar = self.tqdm_cls(**tqdm_kwargs)

    def __call__(self, env: CallbackEnv) -> None:
        super().__call__(env)
        if self.pbar is None:
            return

        # update postfix
        if len(env.evaluation_result_list) > 0:
            # If OrderedDict is not used, the order of display is disjointed and slightly difficult to see.
            # https://github.com/microsoft/LightGBM/blob/a97c444b4cf9d2755bd888911ce65ace1fe13e4b/python-package/lightgbm/callback.py#L56-66
            if self.early_stopping_callback is not None:
                postfix = OrderedDict(
                    [
                        (
                            f"{entry[0]}'s {entry[1]}",
                            f"{entry[2]:g}{'=' if entry[2] == best_score else ('>' if cmp_op else '<')}"
                            f"{best_score:g}@{best_iter}it",
                        )
                        for entry, cmp_op, best_score, best_iter in zip(
                            env.evaluation_result_list,
                            self.early_stopping_callback.cmp_op,
                            self.early_stopping_callback.best_score,
                            self.early_stopping_callback.best_iter,
                        )
                    ]
                )
            else:
                postfix = OrderedDict(
                    [
                        (f"{entry[0]}'s {entry[1]}", f"{entry[2]:g}")
                        for entry in env.evaluation_result_list
                    ]
                )
            self.pbar.set_postfix(ordered_dict=postfix, refresh=False)

        self.pbar.n += 1
        self.pbar.refresh()



from ._base import EstimatorWrapperBase
from ._dart_early_stopping_wrapper import LGBMDartEarlyStoppingSimpleWrapper
from ._early_stopping import LGBMDartEarlyStoppingEstimator, LGBMEarlyStoppingEstimator

__all__ = [
    "LGBMDartEarlyStoppingSimpleWrapper",
    "LGBMEarlyStoppingEstimator",
    "LGBMDartEarlyStoppingEstimator",
    "EstimatorWrapperBase",
]
